/*********************************************************************
 *
 * File:         application.c
 *
 * Description:  NPAC Conferencing System API functions for applications 
 *               management 
 *
 * Author:       Tomasz Stachowiak, NPAC
 *
 * Date:         10-22-96
 *
 *********************************************************************/
#include <stdio.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>

#include <netdb.h>
#include <unistd.h>

#include "common.h"

void ncsReadAppMessage(ncsApplication *app);

int multicast = 1;
int database = 0;

extern int errno;


/************************************************************************** 
 *
 * Send request to the Conference Engine 
 *
 **************************************************************************/
int appEngineRequest(ncsApplication *app, short tag, char *data, int len)
{
  MsgHeader header;

  header.tag = tag;
  cuserid(header.sender.user);
  header.sender.host = gethostid();
  header.dataLength = len;

  if(app->ctrl_socket > 0)
    sendControlMessage(app->ctrl_socket, &header, data, len);

  return 1;
}

/************************************************************************** 
 *
 * Conference Engine messages callbacks
 *
 **************************************************************************/
void appAddMember(ncsApplication *app, char *data)
{
  Participant *newmember;

  newmember = (Participant *)data;

  PRINTF("Application: member added %s@%d handle %d\n", newmember->user, newmember->host, newmember->handle);

  app->memberActive[newmember->handle] = 1;
  memcpy(&app->member[newmember->handle], newmember, sizeof(Participant));

  app->numMembers++;

  if(app->addMbrCallback != NULL)
    app->addMbrCallback(newmember->handle);

  PRINTF("App: member added\n");
}

void appDeleteMember(ncsApplication *app, char *data)
{
  int *no;

  no = (int *)data;
  app->memberActive[*no] = 0;
  app->numMembers--;

  if(app->delMbrCallback != NULL)
    app->delMbrCallback(*no);
}

void appPriorityData(ncsApplication *app, char *data, int len)
{
  AppData *appdata;
  int size;

  appdata = (AppData *)data;
  size = len - sizeof(AppData) + 1;
  if(app->dataCallback != NULL)
    app->dataCallback(appdata->tag, appdata->handle, appdata->data, size);
}

void appQuit(ncsApplication *app)
{
  if((multicast) && (app->localHandle >= 0))
    closeMulticast(app->socket);
  close(app->socket);
  close(app->ctrl_socket);
  free(app);
}

/***********************************************************************
 * 
 * Control messages processing procedure
 *
 ***********************************************************************/
void appProcessMessage(ncsApplication *app, unsigned char sessionID, 
		       unsigned char tag, void *data, int len)
{
  int size;
  InviteData *info;

  info = (InviteData *)data;
  switch((LocalTag)tag)
    {
    case indLocalSessionAddMember:
      PRINTF("Application: add member\n");
      appAddMember(app, data);
      break;

    case indLocalSessionDeleteMember:
      PRINTF("Application: delete member\n");
      appDeleteMember(app, data);
      break;

    case indLocalSessionLocalMember:
      PRINTF("Application: local member %d\n", *(int *)data);
      app->localHandle = *(int *)data;
      app->sessionID = sessionID;
      if(multicast)
	initMulticast(app->socket, sessionID, app->port, &app->mcast_addr);

      if(app->startingCallback != NULL)
	app->startingCallback();
      break;

    case indAppData:
      PRINTF("Application: priority data\n");
      appPriorityData(app, data, len);
      break;

    case indLocalAppQuit:
      PRINTF("Application: quit\n");
      if(app->quitCallback != NULL)
	app->quitCallback();
      else
	{
	  appQuit(app);
	  exit(0);
	}
      break;

    case indDbStart:
      PRINTF("Application: database start\n");
      dbInit(app->port + 2);
      database = 1;
      break;

    case indDbStop:
      PRINTF("Application: database start\n");
      database = 0;
      dbClose();
      break;

    default:
      fprintf(stderr, "Application: wrong message tag %d\n", (LocalTag)tag);
    }
}

/***********************************************************************
 * 
 * Control messages reading procedure
 *
 ***********************************************************************/
void ncsReadAppMessage(ncsApplication *app)
{
  char *msg;
  MsgHeader header;
  int err;

  err = readMsgHeader(app->ctrl_socket, &header);
  if(err < 0)
    {
      fprintf(stderr, "App: data reading error\n");
      return;
    }
  if(err == 0)
    {
      fprintf(stderr, "App: engine killed, exitting...\n");
      if(app->quitCallback != NULL)
	app->quitCallback();
      else
	{
	  appQuit(app);
	  exit(0);
	}
    }
  if(header.dataLength > 0)
    {
      msg = (char *)malloc(header.dataLength);
      err = readAll(app->ctrl_socket, msg, header.dataLength);
      if(err < 0)
	{
	  fprintf(stderr, "App: data reading error\n");
	  return;
	}
      if(err == 0)
	{
	  fprintf(stderr, "App: engine killed, exitting...\n");
	  if(app->quitCallback != NULL)
	    app->quitCallback();
	  else
	    {
	      appQuit(app);
	      exit(0);
	    }
	}
    }
  else
    msg = NULL;
  
  appProcessMessage(app, header.session, header.tag, msg, header.dataLength);
  
  if(msg != NULL)
    free(msg);
}

/************************************************************************** 
 *
 * API functions 
 *
 **************************************************************************/
ncsApplication *ncsAppAttach(int port)
{
  ncsApplication *app;
  long hostid;

  app = malloc(sizeof(ncsApplication));
  
  app->port = port;
  app->localHandle = -1;
  


  /* Initialize application data socket */
  if ((app->socket = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    {
      fprintf(stderr, "Application: can't open socket\n");
      free(app);
      return NULL;
    }
  
  bzero(&app->addr, sizeof(struct sockaddr_in));
  app->addr.sin_family = AF_INET;
  app->addr.sin_addr.s_addr = htonl(INADDR_ANY);
  app->addr.sin_port        = htons(port);
  
  if(bind(app->socket, &app->addr, sizeof(struct sockaddr_in)) < 0)
    {
      PRINTF("Application: can't bind socket\n");
      free(app);
      return NULL;
    }

  /* Initialize application control socket */
  if ((app->ctrl_socket = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
      fprintf(stderr, "Application: can't open control socket\n");
      free(app);
      return NULL;
    }

  /* Set engine address */
  bzero(&app->daemon_addr, sizeof(struct sockaddr_in));
  app->daemon_addr.sin_family = AF_INET;
  hostid = gethostid();
  app->daemon_addr.sin_addr.s_addr = htonl(hostid);
  app->daemon_addr.sin_port        = htons(DAEMON_PORT);

  if(connect(app->ctrl_socket, &app->daemon_addr, sizeof(app->daemon_addr)) < 0)
    { 
      PRINTF("Application: can not connect to engine\n");
      return NULL;
    }
  
  

  /* Set app address */
  bzero(&app->ctrl_addr, sizeof(struct sockaddr_in));
  app->ctrl_addr.sin_family = AF_INET;
  hostid = gethostid();
  app->ctrl_addr.sin_addr.s_addr = htonl(hostid);
  app->ctrl_addr.sin_port        = port;

  PRINTF("Application: attach, port %d\n", app->ctrl_addr.sin_port    );

  /* Send request to engine */
  appEngineRequest(app, reqNewLocalApplicationAttach, 
		   (char *)&app->ctrl_addr.sin_port, sizeof(short));

                        
  app->numMembers = 0;
  app->startingCallback = NULL;
  app->quitCallback = NULL;
  app->addMbrCallback = NULL;
  app->delMbrCallback = NULL;


  return app;
}

void ncsAppDettach(ncsApplication *app)
{
  if(app != NULL)
    {
      /* Send request to engine */
      appEngineRequest(app, reqAppDettach, 
		       (char *)&app->port, sizeof(int));
      appQuit(app);

      if(database)
	{
	  dbClose();
	}
    }
}

int ncsAppSendData(ncsApplication *app, int mbr, int type, void *data, int len)
{
  ncsData *ncsdata;
  int size;
  int i;
  int sent = 0;

  if(len == 0)
    {
      size = sizeof(ncsData);
      ncsdata = malloc(size);
      ncsdata->data[0] = 0;
    }
  else
    {
      size = len + sizeof(ncsData) - 1;
      ncsdata = malloc(size);
      memcpy(ncsdata->data, data, len); 
    }

  ncsdata->type = type;
  ncsdata->handle = app->localHandle;

  if(mbr < 0)  /* send to all members */
    {
      if(multicast)
	{
	  if(sendto(app->socket, ncsdata, size, 
		    0, &app->mcast_addr, sizeof(struct sockaddr_in)) < 0)
	    return 0;  
	}
      else
	{
	  for(i=0; i<MAX_MEMBERS; i++)
	    if(app->memberActive[i])
	      {
		if(i != app->localHandle)
		  {
		    PRINTF("Application: data sent to user %d, size %d\n", i, size);
		    sendPacket(app->socket, app->member[i].host, app->port, ncsdata, size);
		  }
	      }
	}

      if(database)
	{
	  dbAdd(ncsdata, size);
	}
    }
  else   /* send to one specific member */
    {
      if(app->memberActive[mbr])
	if(mbr != app->localHandle)
	  {
	    sendPacket(app->socket, app->member[mbr].host, app->port, ncsdata, size);
	  }

      /* if message to initiator find initiator */
      if((mbr == 0) && (app->memberActive[mbr] == 0))
	{
	  i = 0;
	  while((i < MAX_MEMBERS) && (!sent))
	    {
	      if(i != app->localHandle)
		{
		  PRINTF("Application: data sent to user %d, size %d\n",i, size);
		  sendPacket(app->socket, app->member[i].host, app->port, ncsdata, size);
		  sent = 1;
		}
	      i++;
	    }
	}
    }      

  free(ncsdata);
}

int ncsAppSendPriorityData(ncsApplication *app, int mbr, int type, void *data, int len)
{
  AppData *appdata;
  int size;

  size = sizeof(AppData) + len - 1;
  appdata = malloc(size);
  appdata->handle = app->localHandle;
  appdata->recID = mbr;
  appdata->port = app->port;
  appdata->tag = type;
  memcpy(appdata->data, data, len);

  appEngineRequest(app, reqAppData, appdata, size);

  free(appdata);
}

/*************************************************************************
 * 
 * Proccesses the NCS message incoming from the application control  
 * socket (from the Conference Engine) 
 *
 *************************************************************************/ 
int ncsAppProcessIncoming(ncsApplication *app, fd_set *set)
{
  int msg_len, n, remote_addr_len;
  struct sockaddr_in remote_addr;
  void *data;

  remote_addr_len = sizeof(struct sockaddr_in);
  if(FD_ISSET(app->ctrl_socket, set))
    {
      ncsReadAppMessage(app);
    }
  else
    return 0;

  return 1;
}

/*************************************************************************
 * 
 * Receive the application data(if any) from the application data  
 * socket (from remote applications) 
 *
 *************************************************************************/ 
int ncsAppGetData(ncsApplication *app, fd_set *set, char **data)
{
  int msg_len, n, remote_addr_len;
  struct sockaddr_in remote_addr;

  remote_addr_len = sizeof(struct sockaddr_in);
  if(FD_ISSET(app->socket, set))
    {
      PRINTF("Application: there are data\n");

      *data = malloc(MAX_UDP_SIZE);
      n = recvfrom(app->socket, (char *)*data, MAX_UDP_SIZE, 
		   0, &remote_addr, &remote_addr_len); 
      if (n < 0)
	{
	  fprintf(stderr, "Application: recvfrom error\n", errno);
	  return 0;
	}

      if(database)
	{
	  dbAdd(*data, n);
	}
    }
  else
    return 0;

  return n;
}  


/***********************************************************************
 * 
 * Get the last file descriptor used by NCS and set the file descriptors
 * set 'set' for all descriptors used by NCS
 *
 ***********************************************************************/
int ncsAppGetFD(ncsApplication *app, fd_set *set)
{
  FD_ZERO(set);
  FD_SET(app->socket, set);
  FD_SET(app->ctrl_socket, set);
  return app->ctrl_socket;
}

int ncsAppGetLocalHandle(ncsApplication *app)
{
  return app->localHandle;
}

int ncsAppGetNumMembers(ncsApplication *app)
{
  return app->numMembers;
}

void ncsAppAddCallback(ncsApplication *app, short tag, void (* function))
{
  switch(tag)
    {
    case tagAppStarting:
      app->startingCallback = function;
      break;

    case tagAppQuit:
      app->quitCallback = function;
      break;

    case tagAppData:
      app->dataCallback = function;
      break;

    case tagAppAddMbr:
      app->addMbrCallback = function;
      break;

    case tagAppDelMbr:
      app->delMbrCallback = function;
      break;

    default:
      fprintf(stderr, "Unknown application event tag\n");
    }
}
	













